﻿using System;
using System.IO;
using System.Text;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Threading;
using System.Windows.Forms;
using ArgusLib.Controls;
using ArgusLib.Threading;
using ArgusLib.Threading.Collections;
using FreeImageAPI;
using FreeImageAPI.Metadata;

namespace Imagenator
{
	public partial class MainForm : ArgusLib.Controls.BaseForm
	{
		PatternEngine.Handler patternHandler;
		ISaveFlagInputControl inputControl;
		Thread backgroundWorker;
		SynchronizedStack<object> syncedFileStack;
		ArgusLib.Threading.ThreadState threadState;
		Forms.ImageForm formsImageForm;
		ArgusLib.Action backgroundWorkerFinishedAction;

		public MainForm()
		{
			InitializeComponent();
			this.MinimumClientSize = new Size(575, 300);
			this.lbFiles.AllowedKeyShortcuts = ArgusLib.Controls.ListBox.KeyShortcuts.All;
			this.syncedFileStack = new SynchronizedStack<object>();
			this.threadState = new ArgusLib.Threading.ThreadState();
		}

		static ISaveFlagInputControl GetSaveFlagInputControl(FREE_IMAGE_FORMAT format)
		{
			if (format == FREE_IMAGE_FORMAT.FIF_JPEG)
				return new JpegSaveFlagInputControl();
			return null;
		}

		protected override void OnLoad(EventArgs e)
		{
			this.Text = "Imagenator " + ArgusLib.Environment.GetAssemblyVersion().ToString();

			StringBuilder filter = new StringBuilder();
			StringBuilder allFormats = new StringBuilder();
			Array formats = Enum.GetValues(typeof(FREE_IMAGE_FORMAT));
			foreach (FREE_IMAGE_FORMAT format in formats)
			{
				if (FreeImage.FIFSupportsReading(format) == false)
					continue;
				string extensions = FreeImage.GetFIFExtensionList(format);
				if (string.IsNullOrEmpty(extensions) == true)
					continue;
				extensions = "*." + extensions.Replace(",", ";*.");
				allFormats.Append(extensions);
				allFormats.Append(';');
				filter.Append('|');
				filter.Append(FreeImage.GetFIFDescription(format));
				filter.Append('|');
				filter.Append(extensions);
			}
			allFormats.Remove(allFormats.Length - 1, 1);
			filter.Insert(0, "All Supported Formats|" + allFormats.ToString());
			this.ofdInput.Filter = filter.ToString();
			this.ofdInput.FilterIndex = 0;

			this.patternHandler = PatternEngine.GetHandler(this.tbPattern.Text);

			foreach (PatternEngine.Placeholder placeholder in PatternEngine.Placeholders)
			{
				this.lbVars.Items.Add('{' + placeholder.Identifier + "}: " + placeholder.Description);
			}

			this.formsImageForm = new Forms.ImageForm();

			this.LoadSettings();

			base.OnLoad(e);
		}

		private void LoadSettings()
		{
			Settings settings = Settings.Load<Settings>();
			this.cbFormat.Select(settings.OutputFormat);
			this.cbFilter.Select(settings.ResizingAlgorithm);

			this.tbPattern.Text = settings.OutputPattern;

			if (settings.MainFormBounds.IsEmpty == false)
			{
				Point topLeft = new Point(settings.MainFormBounds.Left, settings.MainFormBounds.Top);
				Screen screen = Screen.FromPoint(topLeft);
				if (screen.WorkingArea.Contains(topLeft) == true)
					this.Bounds = settings.MainFormBounds;
			}

			if (settings.ImageFormBounds.IsEmpty == false)
			{
				Point topLeft = new Point(settings.ImageFormBounds.Left, settings.ImageFormBounds.Top);
				Screen screen = Screen.FromPoint(topLeft);
				if (screen.WorkingArea.Contains(topLeft) == true)
					this.formsImageForm.Bounds = settings.ImageFormBounds;
			}

			this.nudSize.Value = new decimal(settings.OutputSize);

			if (settings.ShowPreviewImages == true)
			{
				this.lbFiles.Appearance = ArgusLib.Controls.ListBox.Appearances.ImageAndText;
				this.lbFiles.ItemHeight = settings.PreviewImageHeight;
				this.threadState.Resume();
			}
			else
			{
				this.lbFiles.Appearance = ArgusLib.Controls.ListBox.Appearances.Text;
				this.lbFiles.ItemHeight = this.lbFiles.Font.Height;
				this.threadState.Suspend();
			}
		}

		private void SaveSettings()
		{
			Settings settings = Settings.Load<Settings>();
			settings.OutputFormat = this.cbFormat.SelectedItem;
			settings.ResizingAlgorithm = this.cbFilter.SelectedItem;
			settings.OutputPattern = this.tbPattern.Text;
			settings.MainFormBounds = this.Bounds;
			settings.OutputSize = (int)this.nudSize.Value;
			settings.ImageFormBounds = this.formsImageForm.Bounds;

			if (this.inputControl != null)
				this.inputControl.Close();

			settings.Save();
		}

		protected override void OnFormClosing(FormClosingEventArgs e)
		{
			if (this.backgroundWorker != null && this.backgroundWorker.IsAlive == true)
			{
				this.backgroundWorkerFinishedAction = new ArgusLib.Action(() => { this.Close(); });
				this.threadState.RequestCancel();
				e.Cancel = true;
				this.Enabled = false;
			}
			this.SaveSettings();
			base.OnFormClosing(e);
		}

		private void BackgroundWorker_Finished()
		{
			this.backgroundWorker = null;
			if (this.backgroundWorkerFinishedAction != null)
				this.backgroundWorkerFinishedAction();
		}

		private void AddFiles(string[] filenames)
		{
			ThreadPool.QueueUserWorkItem(new WaitCallback((arg) =>
				{
					object[] keys = new object[filenames.Length];
					for (int i = 0; i < filenames.Length; i++)
						keys[i] = filenames[i];
					this.lbFiles.Invoke(new ArgusLib.Action(() => { this.lbFiles.Items.AddRange(keys); }));

					if (this.backgroundWorker == null || this.backgroundWorker.IsAlive == false)
					{
						this.backgroundWorker = new Thread(new ThreadStart(() =>
						{
							Parallelization.TaskBody[] tasks = this.AddFilesToBackgroundWorker(filenames);
							Parallelization.ProcessTasks(tasks, this.threadState);
							this.Invoke(new ArgusLib.Action(this.BackgroundWorker_Finished));
						}));
						this.backgroundWorker.Start();
					}
					else
					{
						this.syncedFileStack.PushRange(keys);
					}
				}));
		}

		private Parallelization.TaskBody[] AddFilesToBackgroundWorker(object[] keys)
		{
			Parallelization.TaskBody[] tasks = new Parallelization.TaskBody[keys.Length];
			Settings settings = Settings.Load<Settings>();
			for (int i = 0; i < keys.Length; i++)
			{
				object key = keys[i];
				string filename = (string)key;
				tasks[i] = new Parallelization.TaskBody((threadState) =>
				{
					if (threadState.CancelRequested == true)
						return null;
					if (settings.ShowPreviewImages == true)
					{
						using (FreeImageBitmap fib = new FreeImageBitmap(filename))
						{
							if (threadState.CancelRequested == true)
								return null;
							if (fib.ImageFormat != FREE_IMAGE_FORMAT.FIF_UNKNOWN)
							{
								Bitmap bitmap = fib.ToBitmap();
								if (threadState.CancelRequested == true)
									return null;
								this.lbFiles.Invoke(new ArgusLib.Action(() => { this.lbFiles.Images.Add(key, bitmap); }));
							}
						}
					}
					object[] newKeys = this.syncedFileStack.Empty();;
					return this.AddFilesToBackgroundWorker(newKeys);
				});
			}
			return tasks;
		}

		private void tbOutputPath_DoubleClick(object sender, EventArgs e)
		{
			if (this.fsdOutput.ShowDialog() == System.Windows.Forms.DialogResult.OK)
				this.tbOutputPath.Text = this.fsdOutput.SelectedPath;
		}

		private void cmLbFilesAddFiles_Click(object sender, EventArgs e)
		{
			if (this.ofdInput.ShowDialog() != System.Windows.Forms.DialogResult.OK)
				return;

			this.AddFiles(this.ofdInput.FileNames);
		}

		private void cmLbFilesAddFolders_Click(object sender, EventArgs e)
		{
			if (this.fsdInput.ShowDialog() != System.Windows.Forms.DialogResult.OK)
				return;

			List<string> filenames = new List<string>();
			foreach (string dir in this.fsdInput.SelectedPaths)
			{
				string[] files = Directory.GetFiles(dir, "*", SearchOption.AllDirectories);
				foreach (string file in files)
				{
					if (FreeImage.FIFSupportsReading(FreeImage.GetFIFFromFilename(file)) == true)
						filenames.Add(file);
				}
			}
			this.AddFiles(filenames.ToArray());
		}

		private void lbFiles_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (this.lbFiles.SelectedIndex < 0)
			{
				this.tbPathPreview.Text = null;
				return;
			}

			string imagePath = (string)this.lbFiles.SelectedItem;
			int number = this.lbFiles.SelectedIndex;
			FREE_IMAGE_FORMAT format = this.cbFormat.SelectedItem;

			System.Threading.ThreadPool.QueueUserWorkItem(new WaitCallback((arg) =>
				{
					ImageInfo imageInfo = ImageInfo.GetInfo(imagePath, number);
					if (this.patternHandler != null)
					{
						string path = Path.Combine(this.tbOutputPath.Text, this.patternHandler(imageInfo));
						string ext = FreeImage.GetPrimaryExtensionFromFIF(format);
						if (string.IsNullOrEmpty(ext) == false)
							path += '.' + ext;

						this.Invoke(new ArgusLib.Action(() => { this.tbPathPreview.Text = path; }));
					}

					if (this.formsImageForm.IsShown == true && this.formsImageForm.Visible == true)
					{
						using (FreeImageBitmap fib = new FreeImageBitmap(imagePath))
						{
							this.formsImageForm.BackgroundImage = fib.ToBitmap();
						}
					}
				}));
		}

		private void tbPattern_TextChanged(object sender, EventArgs e)
		{
			this.patternHandler = PatternEngine.GetHandler(this.tbPattern.Text);
		}

		private void lbVars_DoubleClick(object sender, EventArgs e)
		{
			if (this.lbVars.SelectedIndex < 0)
				return;
			string s = (string)this.lbVars.SelectedItem;
			int i = s.IndexOf(':');
			this.tbPattern.Insert(s.Substring(0, i));
		}

		private void cbFormat_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (this.inputControl != null)
				this.inputControl.Close();

			this.panelInputControl.Controls.Clear();
			this.inputControl = MainForm.GetSaveFlagInputControl(this.cbFormat.SelectedItem);
			if (this.inputControl != null)
			{
				Control c = this.inputControl.Control;
				c.Dock = DockStyle.Fill;
				this.panelInputControl.Controls.Add(c);
			}
		}

		private void bStart_Click(object sender, EventArgs e)
		{
			if (this.patternHandler == null)
			{
				MessageBox.Show("Invalid pattern: "+this.tbPattern.Text, "Error");
				return;
			}

			if (Directory.Exists(this.tbOutputPath.Text) == false)
			{
				MessageBox.Show("Invalid output path: "+this.tbOutputPath.Text, "Error");
				return;
			}

			this.threadState.Suspend();

			FREE_IMAGE_FORMAT format = this.cbFormat.SelectedItem;
			string ext = FreeImage.GetPrimaryExtensionFromFIF(format);
			if (string.IsNullOrEmpty(ext) == false)
				ext = '.' + ext;
			FREE_IMAGE_SAVE_FLAGS saveFlags = FREE_IMAGE_SAVE_FLAGS.DEFAULT;
			if (this.inputControl != null)
				saveFlags = this.inputControl.SaveFlags;

			int newSize = (int)this.nudSize.Value;
			FREE_IMAGE_FILTER filter = this.cbFilter.SelectedItem;

			ProgressBox.ShowFor(this.lbFiles.Items.Count, new ArgusLib.Threading.Parallelization.ForBody((i, ts)=>
				{
					ProgressBox.ThreadState threadState = (ProgressBox.ThreadState)ts;
					string file = (string)this.lbFiles.Items[i];
					using (FreeImageBitmap image = new FreeImageBitmap(file))
					{
						
						ImageInfo imageInfo = ImageInfo.GetInfo(image, i, file);
						if (imageInfo == null)
							imageInfo = new ImageInfo(File.GetCreationTime(file), i, file);

						string newExt = ext;
						FREE_IMAGE_FORMAT newFormat = format;
						if (newFormat == FREE_IMAGE_FORMAT.FIF_UNKNOWN)
						{
							newFormat = image.ImageFormat;
							newExt = Path.GetExtension(file);
						}

						string destNoExt = Path.Combine(this.tbOutputPath.Text, this.patternHandler(imageInfo));
						string dest = destNoExt + newExt;

						destNoExt = destNoExt.RemoveChars(Path.GetInvalidPathChars());

						for (int j = 1; j < int.MaxValue; j++)
						{
							if (File.Exists(dest) == false)
								break;
							dest = destNoExt + " (" + j.ToString() + ')' + newExt;
						}

						string dir = Path.GetDirectoryName(dest);
						if (Directory.Exists(dir) == false)
							Directory.CreateDirectory(dir);

						if (newSize > 0 && (image.Width > newSize || image.Height > newSize))
						{
							//Size size = new Size();
							//float ratio = (float)image.Width/image.Height;
							//if (image.Width > image.Height)
							//{
							//	size.Width = newSize;
							//	size.Height = (int)(newSize / ratio);
							//}
							//else
							//{
							//	size.Height = newSize;
							//	size.Width = (int)(newSize * ratio);
							//}

							Size size = new Size(newSize, newSize);
							size = size.ChangeRatio(image.Size.GetRatio(), ArgusLib.Drawing.ChangeRatioOptions.DecreaseAreaSize);
							image.Rescale(size, filter);
						}
						try
						{
							image.Save(dest, newFormat, saveFlags);
							threadState.SendOutput(i.ToString() + ": Completed");
						}
						catch (Exception exception)
						{
							threadState.SendOutput(i.ToString() + ": Exception: "+exception.Message);
						}
					}
				}), false);
			if (Settings.Load<Settings>().ShowPreviewImages == true)
				this.threadState.Resume();
		}

		#region Drag'n'Drop
		private void lbFiles_DragEnter(object sender, DragEventArgs e)
		{
			if (e.AllowedEffect.HasFlag(DragDropEffects.Link) == false || e.Data.GetDataPresent(DataFormats.FileDrop) == false)
			{
				e.Effect = DragDropEffects.None;
			}
			else
			{
				e.Effect = DragDropEffects.Link;
			}
		}

		private void lbFiles_DragDrop(object sender, DragEventArgs e)
		{
			if (e.Effect != DragDropEffects.Link)
				return;

			string[] input = e.Data.GetData(DataFormats.FileDrop) as string[];
			if (input == null)
				return;

			List<string> filenames = new List<string>();
			foreach (string path in input)
			{
				if (Directory.Exists(path) == true)
				{
					string[] files = Directory.GetFiles(path, "*", SearchOption.AllDirectories);
					foreach (string file in files)
					{
						if (FreeImage.FIFSupportsReading(FreeImage.GetFIFFromFilename(file)) == true)
							filenames.Add(file);
					}
				}
				else if (FreeImage.FIFSupportsReading(FreeImage.GetFIFFromFilename(path)) == true)
				{
					filenames.Add(path);
				}
			}
			this.AddFiles(filenames.ToArray());
		}
		#endregion

		private void lbFiles_DoubleClick(object sender, EventArgs e)
		{
			if (this.lbFiles.SelectedIndex < 0)
				return;

			if (this.formsImageForm.IsShown == true && this.formsImageForm.Visible == true)
				return;

			string path = (string)this.lbFiles.SelectedItem;

			System.Threading.ThreadPool.QueueUserWorkItem(new WaitCallback((arg) =>
				{
					using (FreeImageBitmap fib = new FreeImageBitmap(path))
					{
						this.formsImageForm.BackgroundImage = fib.ToBitmap();
					}
					this.Invoke(new ArgusLib.Action(() => { this.formsImageForm.Show(); }));
				}));
		}

		private void tsmiSettings_Click(object sender, EventArgs e)
		{
			this.SaveSettings();
			Forms.SettingsForm settingsForm = new Forms.SettingsForm(this.lbFiles.Font.Height);
			settingsForm.ShowDialog();
			this.LoadSettings();
		}

		private void lbFiles_KeyShortcutSelectedRemoved(object sender, ArgusLib.Controls.ListBox.KeyShortcutSelectedRemovedEventArgs e)
		{
			this.lbFiles.Images.RemoveKeyRange(e.RemovedItems);
		}

		private void tsmiHelpHomepage_Click(object sender, EventArgs e)
		{
			System.Diagnostics.Process.Start("https://imagenator.codeplex.com");
		}

		private void tsmiHelpDocumentation_Click(object sender, EventArgs e)
		{
			System.Diagnostics.Process.Start("https://imagenator.codeplex.com/documentation");
		}
	}
}
